from .base_parsing_step import BaseParsingStep
from . import mixins


# class ParsingStep(mixins.EventletPerformMixin,
class ParsingStep(mixins.MultiThreadPerformMixin,
                  mixins.RequestsDownloadingMixin,
                  mixins.SafePerformMixin,
                  mixins.RepeatRecursivelyOnExceptionMixin,
                  BaseParsingStep):
    """
    Class for parsing data step by step.
    The class represents one parsing step.

    input_data is required on initializing.
    To start parsing, call the "perform" method.

    This class uses multithreading, "requests" module,
    and tries to execute the code again if an exception occurs
    during the execution of the perform method.
    """

    is_save_data_step = False
    next_step_class = None
    countdown_on_exception = 1
    max_repeats_on_exception = None  # None for no limit



import threading
import requests
import queue
import time
import asyncio
import logging

import eventlet


class RequestsDownloadingMixin(object):
    def download(self, *args, **kwargs):
        self.raw_output_data = requests.get(self.input_data)
        return self.raw_output_data


class SafePerformMixin(object):
    def actions_on_exceptions(self, exception, *args, **kwargs):
        raise NotImplementedError

    def perform(self, *args, **kwargs):
        try:
            super().perform(*args, **kwargs)
        except Exception as ex:
            self.actions_on_exceptions(ex, *args, **kwargs)


class MaxRepeatException(Exception):
    pass


class RepeatRecursivelyOnExceptionMixin(object):
    countdown_on_exception = 1
    max_repeats_on_exception = None  # None for no limit

    def actions_on_exceptions(self, exception, *args, **kwargs):
        attempt = getattr(self, "attempt", 0)
        time.sleep(self.countdown_on_exception)
        max_repeats = self.max_repeats_on_exception

        if max_repeats is not None and attempt <= max_repeats:
            attempt += 1
            self.__class__.perform(input_data=self.input_data, attempt=attempt,
                                   *self.input_args, **self.input_kwargs)
        else:
            raise MaxRepeatException()


class MultiThreadPerformMixin(object):
    perform_in_new_thread = True

    def __init__(self, *args, **kwargs):
        if kwargs.get("tasks", None) is None:
            # print("tasks is None")
            kwargs["tasks"] = queue.Queue()
        kwargs["tasks"].closed = False
        return super().__init__(*args, **kwargs)

    def wait_for_complete(self):
        self.tasks.join()
        self.actions_on_complete()

    def stop_all_uncompleted_tasks(self):
        try:
            self.tasks.closed = True

            for task in self.get_uncompleted_tasks():
                task._tstate_lock.release()
                task._stop()
        except Exception as ex:
            logging.error(
                f"An error has ocured on stop_all_uncompleted_tasks: {ex}")

    def actions_on_complete(self):
        pass

    def check_for_complete(self):
        return any([task.is_alive for task in self.tasks])

    def get_uncompleted_tasks(self):
        if self.tasks:
            return [task for task in self.tasks if task.is_alive]
        return []

    def perform_next_step(self, next_step_class, *args, **kwargs):
        kwargs["next_step_class"] = next_step_class

        # before_time = time.time()
        if not self.tasks.closed and next_step_class is not None:
            if next_step_class.perform_in_new_thread:

                is_daemon = getattr(self, "is_daemon", False)
                worker = threading.Thread(target=super().perform_next_step,
                                          args=args, kwargs=kwargs)
                worker.setDaemon(is_daemon)
                # self.tasks.put(worker)
                self.tasks.put(self)
                worker.start()
            else:
                self.tasks.put(self)
                super().perform_next_step(*args, **kwargs)

    def perform(self, *args, **kwargs):
        is_first_task = self.tasks.empty()
        super().perform(*args, **kwargs)
        if not is_first_task:
            # because first task in current thread
            self.tasks.task_done()


class AsyncMixin:
    async def async_get_next_step_class(self, *args, **kwargs):
        # return self.next_step_class
        return await asyncio.to_thread(
            self.get_next_step_class, *args, **kwargs
        )

    async def async_prepare_input_data(self, *args, **kwargs):
        return await asyncio.to_thread(
            self.prepare_input_data, *args, **kwargs
        )

    async def async_input_data_is_valid(self, *args, **kwargs):
        return await asyncio.to_thread(
            self.input_data_is_valid, *args, **kwargs
        )

    async def async_download(self, *args, **kwargs):
        return await asyncio.to_thread(
            self.download, *args, **kwargs
        )

    async def async_process_raw_data(self, *args, **kwargs):
        return await asyncio.to_thread(
            self.process_raw_data, *args, **kwargs
        )

    async def async_input_data_to_output_data(self, *args, **kwargs):
        return await asyncio.to_thread(
            self.input_data_to_output_data, *args, **kwargs
        )

    async def async_process_if_not_valid(self, *args, **kwargs):
        return await asyncio.to_thread(
            self.process_if_not_valid, *args, **kwargs
        )

    async def async_perform_next_step(self, *args, **kwargs):
        return self.perform_next_step(*args, **kwargs)

    async def async_save_data(self, *args, **kwargs):
        return self.save_data(*args, **kwargs)

    async def async_parse(self, *args, **kwargs):
        return self.parse(*args, **kwargs)

    async def async_perform(self, *args, **kwargs):
        return self.perform(*args, **kwargs)


class EventletPerformMixin(object):
    perform_in_new_thread = True

    def __init__(self, *args, **kwargs):
        if kwargs.get("tasks", None) is None:
            kwargs["tasks"] = eventlet.GreenPool(size=1000)
        return super().__init__(*args, **kwargs)

    def wait_for_complete(self):
        self.tasks.waitall()
        self.actions_on_complete()

    def stop_all_uncompleted_tasks(self):
        raise NotImplementedError

    def actions_on_complete(self):
        pass

    def check_for_complete(self):
        return not bool(self.tasks.running())

    def get_uncompleted_tasks(self):
        return self.tasks.coroutines_running

    def perform_next_step(self, next_step_class, *args, **kwargs):
        kwargs["next_step_class"] = next_step_class

        # before_time = time.time()
        if next_step_class is not None:
            self.tasks.spawn_n(super().perform_next_step, *args, **kwargs)



class BaseParsingStep(object):
    """
    Base class for parsing data step by step.
    The class represents one parsing step.

    input_data is required on initializing.
    To start parsing, call the "perform" method.

    input_data_to_output_data method must be overrated on inherited classes.

    For multithreaded use, override the "perform_next_step" method.
    """

    is_save_data_step = False
    next_step_class = None
    perform_in_new_thread = None
    non_inheritable_kwargs = [
        "num_in_list",
    ]

    def __init__(self, input_data, *args, **kwargs):
        """
        Sets the input parameters as the class fields.
        "input_data" parameter is required
        """
        self.input_data = input_data
        for name, value in kwargs.items():
            setattr(self, name, value)

        self.input_args = args
        self.next_step_args = args
        self.input_kwargs = kwargs
        self.next_step_kwargs = {name: value for name, value in kwargs.items()
                                 if name not in self.non_inheritable_kwargs}

    def get_next_step_class(self, *args, **kwargs):
        """
        Returns "next_step" class
        """
        return self.next_step_class

    def input_data_to_output_data(self, *args, **kwargs):
        """
        processes input_data to produce output_data.
        """
        raise NotImplementedError

    def perform_next_step(self, next_step_class, input_data,
                          *next_step_args, **next_step_kwargs):
        """
        The method that starts the next parsing step.
        The code is run by the calling "perform" method.

        If you want to run parsing in multi-threaded mode or through tasks
        (celery, for example) - override this method.
        The "next_step.perform(*args, **kwargs)" method must be run
        on a new thread, process or task in this case.
        """
        if next_step_class is None:
            return None
        next_step = next_step_class(input_data=input_data,
                                    *next_step_args, **next_step_kwargs)
        return next_step.perform()

    def perform(self):
        """
        Performs data parsing.
        """
        self.input_data_to_output_data()
        self.output_data = getattr(self, "output_data", None)

        assert hasattr(self.output_data, "__iter__") or \
            self.output_data is None, \
            "Wrong data type of output data. " \
            "Must be either an iterable object or None."

        if self.output_data is None:
            return

        for num_in_list, data in enumerate(self.output_data, start=1):
            next_step_class = self.get_next_step_class(data=data)
            if next_step_class:
                self.perform_next_step(next_step_class=next_step_class,
                                       input_data=data,
                                       num_in_list=num_in_list,
                                       *self.next_step_args,
                                       **self.next_step_kwargs)



__title__ = "Parsing steps"
__version__ = "0.0.4"
__author__ = "Viacheslav Loievskyi"
__license__ = "Apache Software License"


from .base import BaseParsingStep
from .base_async import BaseAsyncParsingStep
from .simple import SimpleParsingStep
from .simple_async import SimpleAsyncParsingStep



class BaseParsingStep(object):
    """
    Base class for parsing data step by step.
    The class represents one parsing step.

    input_data is required on initializing.
    To start parsing, call the "perform" method.

    For multithreaded use, override the "perform_next_step" method.
    To bypass locks, override the "download" method,
    adding some protection there.
    (This can happen with a large number of requests)
    """

    is_save_data_step = False
    next_step_class = None
    perform_in_new_thread = None
    non_inheritable_kwargs = [
        "num_in_list",
    ]

    def __init__(self, input_data, *args, **kwargs):
        """
        Sets the input parameters as the class fields.
        "input_data" parameter is required
        """
        # print(f"\n\n\n\n\n__init__ {self.__class__.__name__}")
        self.input_data = input_data
        for name, value in kwargs.items():
            setattr(self, name, value)

        self.input_args = args
        self.next_step_args = args
        self.input_kwargs = kwargs
        self.next_step_kwargs = {name: value for name, value in kwargs.items()
                                 if name not in self.non_inheritable_kwargs}

    def get_next_step_class(self, *args, **kwargs):
        """
        Returns "next_step" class
        """
        # print("get_next_step_class")
        return self.next_step_class

    def prepare_input_data(self, *args, **kwargs):
        """
        Prepares input data and modifies it (if needed).
        The initial data is written to self.raw_input_data.
        """
        # print("prepare_input_data")
        self.raw_input_data = self.input_data

    def input_data_is_valid(self, raise_exceptions=False, *args, **kwargs):
        """
        Input data validation.
        Returns True if input_data is valid,
        or False otherwise
        """
        # print("input_data_is_valid")
        self.validated_input_data = getattr(self, "input_data", None)
        return True

    def download(self, *args, **kwargs):
        """
        Downloading or something else.
        makes "raw_output_data" from "input_data".
        """
        # print("download")
        self.raw_output_data = self.input_data

    def process_raw_data(self, *args, **kwargs):
        """
        Processes input data.
        Takes raw data and writes to the field "output_data"
        """
        # print("process_raw_data")
        self.output_data = [self.raw_output_data]

    def input_data_to_output_data(self, *args, **kwargs):
        """
        processes input_data to produce output_data.
        """
        # print("input_data_to_output_data")
        self.download(*args, **kwargs)
        self.process_raw_data(*args, **kwargs)

    def process_if_not_valid(self, *args, **kwargs):
        """
        Actions that will run in case of invalid "input_data".
        Must be overridden if validation is overridden.
        """
        # print("process_if_not_valid")
        raise NotImplementedError

    def perform_next_step(self, next_step_class, input_data,
                          *next_step_args, **next_step_kwargs):
        """
        The method that starts the next parsing step.
        The code is run by the calling "perform" method.

        If you want to run parsing in multi-threaded mode or through tasks
        (celery, for example) - override this method.
        The "next_step.perform(*args, **kwargs)" method must be run
        on a new thread, process or task in this case.
        """
        # print("perform_next_step")
        if next_step_class is None:
            return None
        next_step = next_step_class(input_data=input_data,
                                    *next_step_args, **next_step_kwargs)
        return next_step.perform()

    def save_data(self, *args, **kwargs):
        """
        This method is called if the "is_save_data_step"
        attribute is set to True.
        Should be overridden in this case.
        """
        # print("save_data")
        raise NotImplementedError

    def parse(self, *args, **kwargs):
        """
        The method is called if
        the "is_save_data_step" attribute is set to False.
        """
        # print("parse")
        self.input_data_to_output_data(*args, **kwargs)

    def perform(self):
        """
        Performs data parsing.
        """
        # print("perform")
        args = self.input_args
        kwargs = self.input_kwargs

        # print(f"args: {args}")
        # print(f"kwargs: {kwargs}")
        self.prepare_input_data(*args, **kwargs)
        if self.input_data_is_valid(*args, **kwargs):
            if self.is_save_data_step:
                self.save_data(*args, **kwargs)
            else:
                self.input_data_to_output_data(*args, **kwargs)

            self.output_data = getattr(self, "output_data", None)
            if type(self.output_data) is not list:
                self.output_data = [self.output_data]
            for num_in_list, data in enumerate(self.output_data, start=1):
                next_step_class = self.get_next_step_class(data=data)
                self.perform_next_step(next_step_class=next_step_class,
                                       input_data=data,
                                       num_in_list=num_in_list,
                                       *self.next_step_args,
                                       **self.next_step_kwargs)
                # return None
        else:
            self.process_if_not_valid(*args, **kwargs)



import time

from parsing_steps import SimpleParsingStep


class ParseMainPage(SimpleParsingStep):
    """
    input_data is main page url.
    input_data = "https://domain.com/"
    """

    def download(self, url):
        time.sleep(0.1)
        return '<html>...<a href="https://domain.com/category/15/"></a>...</html>'

    def get_category_urls(self, html_data):
        """Plug"""
        return [f"https://domain/category/{num}/" for num in range(15, 16)]

    def parse(self, input_data):
        url = input_data
        html_data = self.download(url)

        for category_url in self.get_category_urls(html_data):
            yield ParseListStep({"url": category_url})


class ParseListStep(SimpleParsingStep):
    """
    input_data from previous step.
    input_data = {
        "url": "https://domain.com/category/<id>/"
    }
    """

    def download(self, url):
        time.sleep(0.1)
        if "?page=2" in url:
            return "<html>...</html>"
        return '<html>...<a href="https://domain.com/category/15/?page=2"></a>...</html>'

    def get_item_urls(self, html_data):
        """Plug"""
        return [f"https://domain/category/15/item/{num}/" for num in range(10)]

    def exists_next_page(self, html_data):
        return "?page=" in html_data

    def parse(self, input_data):
        print(f"category_url: {input_data.get('url')}")
        #  here you get data from html, through bs4 or something else.
        html_data = self.download(input_data.get("url"))

        for item_url in self.get_item_urls(html_data):
            yield ParseDetailsStep(
                input_data={"url": item_url},
                inherited_data={  # optional part
                    "category_id": 15,
                    "page": 1 if self.exists_next_page(html_data) else 2
                }
            )

        if self.exists_next_page(html_data):
            # get this data from html
            next_page_url = "https://domain.com/category/15/?page=2"
            yield ParseListStep(
                input_data={"url": next_page_url}
            )


class ParseDetailsStep(SimpleParsingStep):
    """
    input_data from previous step.
    input_data = {
        "url": "https://domain/category/15/item/{num}/"
    }
    """

    def download_and_format(self, url):
        """plug"""
        time.sleep(0.2)
        num = int(list(filter(None, str(url).split("/")))[-1])  # returns num
        page = int(self.inherited_data.get("page"))
        return {
            "id": page*10 + int(num),
            "name": f"product_{page*10 + num}",
            "price": num*10,
        }

    def parse(self, input_data):
        product_data = self.download_and_format(input_data.get("url"))
        yield SaveStep(product_data)


class SaveStep(SimpleParsingStep):
    """
    input_data from previous step.
    input_data = {
        "id": 12345,
        "name": "product",
        "price": 16.50,
    }
    """

    def save_to_db(self, data_as_dict):
        self.inherited_data = self.inherited_data or dict()
        data_as_dict.update(self.inherited_data)
        print(data_as_dict)

    def parse(self, input_data):
        self.save_to_db(input_data)


if __name__ == "__main__":
    first_step = ParseMainPage(input_data="https://domain.com/")
    first_step.perform()  # start scraping



#!/usr/bin/env python

from setuptools import find_packages, setup


def read_file(file_path):
    return open(file_path, "r", encoding="utf-8").read()


setup(
    name="parsing_steps",
    version="0.0.4",
    url="https://github.com/loievskyi/parsing_steps",
    license="Apache Software License",
    description="A simple helper for data parsing.",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    author="Viacheslav Loievskyi",
    author_email="loievskyi.slava@gmail.com",
    packages=find_packages(exclude=["tests*", "parsing_steps/future*", "old_code/*"]),
    include_package_data=True,
    python_requires=">=3.8",
    zip_safe=False,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Internet",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
    ],
    project_urls={
        "Source": "https://github.com/loievskyi/parsing_steps",
    },
)



from typing import Iterable


class SimpleParsingStep:
    """
    Simpliify class for parsing data step by step.
    The class represents one parsing step.

    input_data is required on initializing.
    To start parsing, call the "perform" method.

    "parse" method must be overrated on inherited classes.

    For multithreaded use, override the "perform_next_step" method.
    """

    next_step_class = None

    def __init__(self, input_data, inherited_data=None, **kwargs):
        """
        Sets the input parameters as the class fields.
        "input_data" parameter is required
        """
        for name, value in kwargs.items():
            setattr(self, name, value)
        self.input_data = input_data
        self.inherited_data = inherited_data

    def get_next_step_class(self, next_step_data):
        """
        Returns "next_step" class
        """
        return self.next_step_class

    def parse(self, input_data) -> Iterable:
        """
        Processes input_data to produce output_data. output_data is iterable
        object (list, set, or generator) of parsing steps instances.
        output_data must be returned.
        """
        raise NotImplementedError

    def perform_next_step(self, next_step):
        """
        The method that starts the next parsing step.
        The code is run by the calling "perform" method.

        If you want to run parsing in multi-threaded mode or through tasks
        (celery, for example) - override this method.
        The "next_step.perform(*args, **kwargs)" method must be run
        on a new thread, process or task in this case.
        """
        if next_step is None:
            return None

        if next_step.inherited_data is None:
            next_step.inherited_data = self.inherited_data

        return next_step.perform()

    def perform(self):
        """
        Performs data parsing.
        """
        self.output_data = self.parse(self.input_data)

        assert hasattr(self.output_data, "__iter__") or \
            self.output_data is None, \
            "Wrong data type of output data. " \
            "Must be either an iterable object or None."

        if self.output_data is None:
            return

        for num, next_step in enumerate(self.output_data, start=1):
            if next_step and self.perform_next_step:
                self.perform_next_step(next_step=next_step)
